use std::{
    fs::File,
    io::{BufRead, BufReader, Cursor},
    path::PathBuf,
};

use clap::{crate_authors, crate_version, Parser};
use miette::{bail, miette, IntoDiagnostic, WrapErr};
use syntect::{
    easy::HighlightLines,
    highlighting::{Color, Style, ThemeSet},
    parsing::SyntaxSet,
};

#[derive(Parser)]
#[clap(
    version = crate_version!(),
    author = crate_authors!(),
)]
struct Opt {
    #[clap(subcommand)]
    cmd: Cmd,
}

#[derive(Parser)]
enum Cmd {
    #[clap(about = "Print available themes and exit")]
    ListThemes,

    #[clap(about = "Print available syntaxes and exit")]
    ListSyntaxes,

    #[clap(about = "Highlight a file", alias = "hl")]
    Highlight {
        #[clap(about = "The file to highlight")]
        file: PathBuf,

        #[clap(
            long,
            short,
            default_value = "dracula",
            about = "The color scheme to use"
        )]
        theme: String,

        #[clap(
            long,
            short,
            about = "The syntax to use. Uses file extension if not provided."
        )]
        syntax: Option<String>,
    },
}

fn main() -> miette::Result<()> {
    let opt = Opt::parse();

    let mut ts = ThemeSet::load_defaults();
    ts.themes.insert(
        "dracula".to_string(),
        // TODO: do expensive tmTheme parsing in build script
        ThemeSet::load_from_reader(&mut Cursor::new(include_bytes!(
            "../assets/dracula_theme/Dracula.tmTheme"
        )))
        .into_diagnostic()
        .wrap_err("Failed to load static dracula theme")?,
    );
    let ss = SyntaxSet::load_defaults_newlines();

    match opt.cmd {
        Cmd::ListThemes => {
            for theme in ts.themes.keys() {
                println!("{}", theme);
            }

            Ok(())
        },

        Cmd::ListSyntaxes => {
            let mut syntaxes = ss.syntaxes().to_vec();
            syntaxes.sort_by(|a, b| a.name.cmp(&b.name));
            for syn in syntaxes {
                println!("{}\t[{}]", syn.name, syn.file_extensions.join(", "))
            }

            Ok(())
        },

        Cmd::Highlight {
            file,
            theme,
            syntax,
        } => {
            let theme = if let Some(t) = ts.themes.get(&theme) {
                t
            } else {
                bail!("No such theme! Use list-themes to list themes.");
            };

            let syntax = if let Some(syn) = syntax {
                ss.find_syntax_by_name(&syn)
            } else {
                ss.find_syntax_for_file(&file)
                    .into_diagnostic()
                    .wrap_err("Failed to open file for finding syntax.")?
            };

            let reader = BufReader::new(
                File::open(file)
                    .into_diagnostic()
                    .wrap_err("Failed to open file")?,
            );

            let syntax = syntax.ok_or_else(|| {
                miette!("Couldn't find syntax. Set one that exists.\nList with list-syntaxes.")
            })?;

            let mut hl = HighlightLines::new(syntax, theme);

            print!("<pre>");
            for line in reader.lines() {
                let line = line.into_diagnostic().wrap_err("Failed to read file")?;
                let regions = hl.highlight(&line, &ss);

                for (
                    Style {
                        foreground: Color { r, g, b, .. },
                        ..
                    },
                    txt,
                ) in regions
                {
                    print!(
                        r##"<font color="#{:02x}{:02x}{:02x}">{}</font>"##,
                        r,
                        g,
                        b,
                        html_escape::encode_text(txt)
                    );
                }

                println!();
            }
            println!("</pre>");

            Ok(())
        },
    }
}
